"""Define the structure of a Pynecone app."""

import os
import pickle
from collections import defaultdict
from typing import Any, Callable, Coroutine, Dict, List, Optional, Type, Union

import fastapi
from fastapi.middleware import cors

from pynecone import constants, utils
from pynecone.base import Base
from pynecone.compiler import compiler
from pynecone.compiler import utils as compiler_utils
from pynecone.components.base import Head, Title
from pynecone.components.component import Component, ComponentStyle
from pynecone.event import Event
from pynecone.middleware import HydrateMiddleware, Middleware, LoggingMiddleware
from pynecone.model import Model
from pynecone.state import DefaultState, Delta, State
from pynecone.style import Style

# Define custom types.
ComponentCallable = Callable[[], Component]
Reducer = Callable[[Event], Coroutine[Any, Any, Union[State, Delta]]]


class App(Base):
    """A Pynecone application."""

    # A map from the path to the page.
    pages: Dict[str, Component] = {}

    # A list of urls to stylesheets to include in the app.
    stylesheets: List[str] = []

    # The backend api object.
    api: fastapi.FastAPI = None  # type: ignore

    # The state class to use for the app.
    state: Type[State] = DefaultState

    # The map from the client session token to the state.
    states: Dict[str, State] = {}

    # The styling to apply to each component.
    style: ComponentStyle = {}

    # Middleware to add to the app.
    middleware: List[Middleware] = [HydrateMiddleware()]

    class Config:
        """Config to satisfy pydantic type checking."""

        arbitrary_types_allowed = True

    def __init__(self, *args, **kwargs):
        """Initialize the app.

        Args:
            *args: Args to initialize the app with.
            **kwargs: Kwargs to initialize the app with.
        """
        super().__init__(*args, **kwargs)
        self.states = defaultdict(self.state)
        self.api = fastapi.FastAPI()
        self.add_cors()
        self.add_default_endpoints()

    def add_default_endpoints(self):
        """Add the default endpoints."""
        # To test the server.
        self.get(constants.Endpoint.PING)(_ping)

        # To make state changes.
        self.post(constants.Endpoint.EVENT)(_event(app=self))

    def add_cors(self):
        """Add CORS middleware to the app."""
        self.api.add_middleware(
            cors.CORSMiddleware,
            allow_origins=["*"],
            allow_methods=["*"],
            allow_headers=["*"],
        )

    def get(self, path: Union[str, constants.Endpoint], *args, **kwargs) -> Callable:
        """Register a get request.

        Args:
            path: The endpoint path to link to the request.
            *args: Args to pass to the request.
            **kwargs: Kwargs to pass to the request.

        Returns:
            A decorator to handle the request.
        """
        return self.api.get(str(path), *args, **kwargs)

    def post(self, path: Union[str, constants.Endpoint], *args, **kwargs) -> Callable:
        """Register a post request.

        Args:
            path: The endpoint path to link to the request.
            *args: Args to pass to the request.
            **kwargs: Kwargs to pass to the request.

        Returns:
            A decorator to handle the request.
        """
        return self.api.post(str(path), *args, **kwargs)

    def preprocess(self, state: State, event: Event) -> Optional[Delta]:
        """Preprocess the event.

        Args:
            state: The state to preprocess.
            event: The event to preprocess.

        Returns:
            An optional state to return.
        """
        for middleware in self.middleware:
            out = middleware.preprocess(app=self, state=state, event=event)
            if out is not None:
                return out

    def postprocess(self, state: State, event: Event, delta: Delta) -> Optional[Delta]:
        """Postprocess the event.

        Args:
            state: The state to postprocess.
            event: The event to postprocess.
            delta: The delta to postprocess.

        Returns:
            An optional state to return.
        """
        for middleware in self.middleware:
            out = middleware.postprocess(
                app=self, state=state, event=event, delta=delta
            )
            if out is not None:
                return out

    def add_page(
        self,
        component: Union[Component, ComponentCallable],
        path: Optional[str] = None,
        title: str = "My Pynecone App",
    ):
        """Add a page to the app.

        Args:
            path: The path to add the page to.
            component: The component to display at the page.
            title: The title of the page.
        """
        # If the path is not set, get it from the callable.
        if path is None:
            assert isinstance(
                component, Callable
            ), "Path must be set if component is not a callable."
            path = utils.to_snake_case(component.__name__).replace("_", "-")

        # Generate the component if it is a callable.
        component = component if isinstance(component, Component) else component()

        # Format the path.
        path = path.strip(os.path.sep)
        if path == "":
            path = constants.INDEX

        # Add the title to the component.
        component.children.append(Head.create(Title.create(title)))

        self.pages[path] = component

    def compile(self):
        """Compile the app and output it to the pages folder."""
        # Get the env mode.
        config = utils.get_config()
        if config.ENV != constants.Env.DEV.value:
            print("Skipping compilation in non-dev mode.")
            return

        # Create the models.
        Model.create_all()

        self.pages[constants.DOCUMENT_ROOT] = compiler_utils.create_document_root(
            self.stylesheets
        )
        self.pages[constants.THEME] = {  # type: ignore
            "styles": {
                "global": Style(
                    {k: v for k, v in self.style.items() if not isinstance(k, type)}
                )
            },
        }
        for path, component in self.pages.items():
            # Get the path for the output file.
            output_path = os.path.join(constants.WEB_PAGES_DIR, path + constants.JS_EXT)

            if path == constants.DOCUMENT_ROOT:
                code = compiler.compile_document_root(component)
            elif path == constants.THEME:
                output_path = os.path.join(
                    constants.WEB_UTILS_DIR, path + constants.JS_EXT
                )
                code = compiler.compile_theme(component)  # type: ignore
            else:
                # Add the style and state to the component.
                component.add_style(self.style).set_state(self.state)

                code = compiler.compile_component(
                    component=component,
                    state=self.state,
                )

            utils.mkdir(os.path.dirname(output_path))
            with open(output_path, "w") as f:
                f.write(code)


async def _ping() -> str:
    """Test API endpoint.

    Returns:
        The response.
    """
    return "pong"


def get_redis():
    """Get the redis client.

    Returns:
        The redis client.
    """
    try:
        config = utils.get_config()
        import redis

        redis_host, redis_port = config.REDIS_HOST.split(":")
        print("Using redis at", config.REDIS_HOST)
        return redis.Redis(host=redis_host, port=redis_port, db=0)
    except:
        return None


r = get_redis()


def get_state(app, event):
    """Get the state for an event.

    Args:
        app: The app to get the state for.
        event: The event to get the state for.

    Returns:
        The state for the event.
    """
    if r is None:
        return app.states[event.token]
    redis_state = r.get(event.token)
    if redis_state is not None:
        return pickle.loads(redis_state)
    return app.state()


def set_state(state, event):
    """Set the state for the event.

    Args:
        state: The state to set.
        event: The event to set the state for.

    Returns:
        The state for the event.
    """
    if r is None:
        return
    r.set(event.token, pickle.dumps(state), ex=60 * 60)


def _event(app: App) -> Reducer:
    """Create an event reducer to modify the state.

    Args:
        app: The app to modify the state of.

    Returns:
        A handler that takes in an event and modifies the state.
    """

    async def process(event: Event) -> Delta:
        # Get the state for the session.
        state = get_state(app, event)

        # Preprocess the event.
        pre = app.preprocess(state, event)
        if pre is not None:
            set_state(pre["state"], event)
            return pre

        # Apply the event to the state.
        delta = state.process(event)

        # Postprocess the event.
        post = app.postprocess(state, event, delta)
        if post is not None:
            set_state(post["state"], event)
            return post

        # Return the delta.
        set_state(state, event)
        return delta

    return process
