"""Basic component definitions."""

from __future__ import annotations

import inspect
import os
from abc import ABC
from collections import defaultdict
from typing import _GenericAlias  # type: ignore
from typing import Any, Dict, List, Optional, Set, Tuple, Type, Union

from pynecone import utils
from pynecone.base import Base
from pynecone.components.tags import Tag
from pynecone.event import EventChain, EventHandler, EventSpec
from pynecone.state import State
from pynecone.style import Style
from pynecone.var import Var

ImportDict = Dict[str, Set[str]]


class Component(Base, ABC):
    """The base class for all other pynecone components."""

    # The children nested within the component.
    children: List[Component] = []

    # An optional condition that must be True for the component to render.
    condition: Var[bool]

    # The style of the component.
    style: Style = Style()

    # A dictionary of events bound to the component.
    events: Dict[str, EventChain] = {}

    # The library that the component is based on.
    library: Optional[str] = None

    # The tag to use when rendering the component.
    tag: Optional[str] = None

    # A unique key for the component.
    key: Any = None

    # The app state.
    state: Type[State] = State

    class Config:
        """Pydantic config."""

        arbitrary_types_allowed = True

    @classmethod
    def __init_subclass__(cls, **kwargs):
        """Set default properties.

        Args:
            **kwargs: The kwargs to pass to the superclass.
        """
        super().__init_subclass__(**kwargs)

        for field in cls.get_fields().values():
            # Only set for var types.
            if isinstance(field.type_, type) and issubclass(
                field.type_, (Var, EventSpec, EventChain)
            ):
                field.required = False
                field.default = Var.create(field.default)

    def __init__(self, *args, **kwargs):
        """Initialize the component.

        Args:
            *args: Args to initialize the component.
            **kwargs: Kwargs to initialize the component.
        """
        # Find the style attributes.
        if "style" not in kwargs:
            kwargs["style"] = {}
        else:
            kwargs["style"] = kwargs["style"].copy()

        # Add the events.
        if "events" not in kwargs:
            kwargs["events"] = {}
        else:
            kwargs["events"] = kwargs["events"].copy()

        # Convert cond to condition.
        if "cond" in kwargs:
            kwargs["condition"] = kwargs.pop("cond")

        # Convert other fields to properties.
        fields = self.get_fields()
        for key, value in kwargs.items():
            if key == "events":
                continue
            if key in SYNTHETIC_EVENTS:
                field_type = EventChain
            else:
                # If the key is not in the fields, skip it.
                if key not in fields:
                    continue

                # If the field does not have a type, skip it.
                if not inspect.isclass(fields[key].type_):
                    continue

                field_type = fields[key].type_

            # Convert event handlers into event specs.
            if issubclass(field_type, EventChain) and not isinstance(value, EventChain):
                if not isinstance(value, utils.get_args(Union[List, Set, Tuple])):
                    value = [value]
                value = [
                    event() if isinstance(event, EventHandler) else event
                    for event in value
                ]
                kwargs[key] = EventChain(events=value)

            # Check whether the key is a component prop.
            if issubclass(field_type, Var):
                # Convert any constants into properties.
                kwargs[key] = Var.create(value)

                # Make sure the var types match.
                expected_type = fields[key].outer_type_.__args__[0]
                if isinstance(expected_type, _GenericAlias):
                    expected_type = expected_type.__origin__
                assert issubclass(
                    kwargs[key].type_, expected_type
                ), f"Invalid var passed for {key}, expected {expected_type}, got {kwargs[key].type_}."

            # If the key is an event, add it to the events dict.
            if key in SYNTHETIC_EVENTS:
                kwargs["events"][key] = kwargs[key]
                continue

        for key in kwargs["events"]:
            del kwargs[key]

        kwargs["style"].update(
            {
                attr: value
                for attr, value in kwargs.items()
                if attr not in self.get_fields().keys()
            }
        )
        kwargs["style"] = Style(kwargs["style"])
        super().__init__(*args, **kwargs)

    def __repr__(self) -> str:
        """Represent the component in React.

        Returns:
            The code to render the component.
        """
        return self.render()

    def __str__(self) -> str:
        """Represent the component in React.

        Returns:
            The code to render the component.
        """
        return self.render()

    def _render(self) -> Tag:
        """Define how to render the component in React.

        Returns:
            The tag to render.
        """
        return Tag(name=self.tag).add_attrs(
            **{
                attr: getattr(self, attr)
                for attr in set(self.get_fields()) - set(Component.get_fields())
            }
        )

    @classmethod
    def create(cls, *children, **attrs) -> Component:
        """Create the component.

        Args:
            *children: The children of the component.
            **attrs: The attributes of the component.

        Returns:
            The component.
        """
        from pynecone.components.base.bare import Bare

        children = [
            Bare.create(contents=Var.create(child))
            if not isinstance(child, Component)
            else child
            for child in children
        ]
        return cls(children=children, **attrs)

    def _add_style(self, style):
        self.style.update(style)

    def add_style(self, style: ComponentStyle) -> Component:
        """Add additional style to the component and its children.

        Args:
            style: A dict from component to styling.

        Returns:
            The component with the additional style.
        """
        if type(self) in style:
            component_style = Style(style[type(self)])
            component_style = {
                k: v for k, v in component_style.items() if k not in self.style
            }
            self._add_style(component_style)
        for child in self.children:
            child.add_style(style)
        return self

    def set_state(self, state: Type[State]) -> Component:
        """Set the state of the component.

        Args:
            state: The state to set.

        Returns:
            The component with the state set.
        """
        self.state = state
        for child in self.children:
            child.set_state(state)
        return self

    def render(self) -> str:
        """Render the component.

        Returns:
            The code to render the component.
        """
        tag = self._render()
        return str(
            tag.add_attrs(**self.events, key=self.key, sx=self.style).set(
                condition=self.condition,
                contents=os.linesep.join(
                    [str(tag.contents)] + [child.render() for child in self.children]
                ),
            )
        )

    def _get_custom_code(self) -> str:
        """Get custom code for the component.

        Returns:
            The custom code.
        """
        return ""

    def get_custom_code(self) -> str:
        """Get custom code for the component and its children.

        Returns:
            The custom code.
        """
        code = self._get_custom_code()
        for child in self.children:
            child_code = child.get_custom_code()
            if child_code != "" and child_code not in code:
                code += child_code
        return code

    def _get_imports(self) -> ImportDict:
        if self.library is not None and self.tag is not None:
            return {self.library: {self.tag}}
        return {}

    def get_imports(self) -> ImportDict:
        """Get all the libraries and fields that are used by the component.

        Returns:
            The import dict with the required imports.
        """
        return merge_imports(
            self._get_imports(), *[child.get_imports() for child in self.children]
        )


def merge_imports(*imports) -> ImportDict:
    """Merge two import dicts together.

    Args:
        *imports: The list of import dicts to merge.

    Returns:
        The merged import dicts.
    """
    all_imports = defaultdict(set)
    for import_dict in imports:
        for lib, fields in import_dict.items():
            for field in fields:
                all_imports[lib].add(field)
    return all_imports


# Map from component to styling.
ComponentStyle = Dict[Union[str, Type[Component]], Any]

SYNTHETIC_EVENTS = {
    "on_copy",
    "on_cut",
    "on_paste",
    "on_composition_end",
    "on_composition_start",
    "on_composition_update",
    "on_key_down",
    "on_key_press",
    "on_key_up",
    "on_focus",
    "on_blur",
    "on_change",
    "on_input",
    "on_invalid",
    "on_submit",
    "on_error",
    "on_load",
    "on_click",
    "on_context_menu",
    "on_double_click",
    "on_drag",
    "on_drag_end",
    "on_drag_enter",
    "on_drag_exit",
    "on_drag_leave",
    "on_drag_over",
    "on_drag_start",
    "on_drop",
    "on_mouse_down",
    "on_mouse_enter",
    "on_mouse_leave",
    "on_mouse_move",
    "on_mouse_out",
    "on_mouse_over",
    "on_mouse_up",
    "on_pointer_down",
    "on_pointer_move",
    "on_pointer_up",
    "on_pointer_cancel",
    "on_got_pointer_capture",
    "on_lost_pointer_capture",
    "on_pointer_enter",
    "on_pointer_leave",
    "on_pointer_over",
    "on_pointer_out",
    "on_select",
    "on_touch_cancel",
    "on_touch_end",
    "on_touch_move",
    "on_touch_start",
    "on_scroll",
    "on_wheel",
    "on_load",
    "on_error",
    "on_toggle",
}
