"""A select component."""

from typing import Any, Tuple

from pynecone import utils
from pynecone.components.libs.chakra import ChakraComponent
from pynecone.components.mixins import ControlledMixin
from pynecone.components.tags import Tag
from pynecone.components.typography.text import Text
from pynecone.var import Var


class Select(ChakraComponent, ControlledMixin):
    """Select component is a component that allows users pick a value from predefined options. Ideally, it should be used when there are more than 5 options, otherwise you might consider using a radio group instead."""

    tag = "Select"

    # State var to bind the the input.
    value: Var[str]

    # The placeholder text.
    placeholder: Var[str]

    # The type of input.
    type_: Var[str] = "text"  # type: ignore

    # The border color when the select is invalid.
    error_border_color: Var[str]

    # The border color when the select is focused.
    focus_border_color: Var[str]

    # If true, the select will be disabled.
    is_disabled: Var[bool]

    # If true, the form control will be invalid. This has 2 side effects: - The FormLabel and FormErrorIcon will have `data-invalid` set to true - The form element (e.g, Input) will have `aria-invalid` set to true
    is_invalid: Var[bool]

    # If true, the form control will be readonly
    is_read_only: Var[bool]

    # If true, the form control will be required. This has 2 side effects: - The FormLabel will show a required indicator - The form element (e.g, Input) will have `aria-required` set to true
    is_required: Var[bool]

    # "outline" | "filled" | "flushed" | "unstyled"
    variant: Var[str]

    def _get_local_args(self) -> Tuple[str, ...]:
        return ("e",)

    def _get_handler_args(self) -> Tuple[Tuple[str, Any], ...]:
        return ((self.value.name, "e.target.value"),)

    @property
    def default_placeholder(self) -> str:
        """Get the default placeholder text.

        Returns:
            The default placeholder text.
        """
        return utils.to_title(self.value.name)

    def _render(self) -> Tag:
        return (
            super()
            ._render()
            .add_attrs(
                placeholder=self.placeholder
                if self.placeholder is not None
                else self.default_placeholder,
                type=self.type_,
            )
        )


class Option(Text):
    """A button component."""

    tag = "option"

    value: Var[int]
