"""Mixin for controlled components."""

from abc import ABC
from typing import Any, Optional, Tuple

from pynecone.components.component import Component
from pynecone.components.tags import Tag
from pynecone.event import EventChain, EventSpec
from pynecone.var import BaseVar, Var


class ControlledMixin(Component, ABC):
    """A mixin for components that are bound to a state."""

    # State var to bind the the input.
    value: Var[Any]

    class Config:
        """Pydantic config."""

        arbitrary_types_allowed = True

    def _get_controlled_var(self) -> Var:
        """Get the controlled var.

        Returns:
            The controlled var.
        """
        return self.value

    def _get_handler(self) -> Optional[EventChain]:
        """Get the handler to control the value.

        Returns:
            The handler to control the value.
        """
        if "on_change" in self.events:
            return self.events["on_change"]
        return None

    def _get_local_args(self) -> Tuple[str, ...]:
        """Get the local arguments for the component.

        Returns:
            The local arguments for the component.
        """
        return ("e",)

    def _get_handler_args(self) -> Tuple[Tuple[str, Any], ...]:
        """Get the args to pass to the handler.

        Returns:
            The args to pass to the handler.
        """
        prop = self._get_controlled_var()
        return ((prop.name, "e.target.value"),)

    def _get_on_change(self) -> Optional[EventChain]:
        prop = self._get_controlled_var()
        handler = self._get_handler()
        if prop is None:
            return None
        if handler is None and isinstance(prop, BaseVar):
            on_change = self.state.get_class_var(prop.get_setter_name().split("."))
        else:
            on_change = handler
        if on_change is None:
            return None
        if isinstance(on_change, EventChain):
            return EventChain(
                events=[
                    EventSpec(
                        handler=e.handler,
                        local_args=self._get_local_args(),
                        args=self._get_handler_args(),
                    )
                    for e in on_change.events
                ]
            )
        spec = EventSpec(
            handler=on_change,
            local_args=self._get_local_args(),
            args=self._get_handler_args(),
        )
        return EventChain(events=[spec])

    def _render(self) -> Tag:
        on_change = self._get_on_change()
        if on_change is not None:
            self.events["on_change"] = on_change
        return super()._render()
