"""Pynecone CLI to create, run, and deploy apps."""

import os

import requests
import typer

from pynecone import constants, utils
from pynecone.compiler import templates

# Create the app.
cli = typer.Typer()


@cli.command()
def init():
    """Initialize a new Pynecone app."""
    # The app name is the same as the name of the current directory.
    app_name = os.getcwd().split(os.path.sep)[-1]
    typer.echo(f"Initializing {app_name}")

    # Only create the app directory if it doesn't exist.
    if not os.path.exists(constants.CONFIG_FILE):
        # Create a configuration file.
        with open(constants.CONFIG_FILE, "w") as f:
            f.write(templates.PCCONFIG.format(app_name=app_name))

        # Initialize the app directory.
        utils.cp(constants.APP_TEMPLATE_DIR, app_name)
        utils.mv(
            os.path.join(app_name, constants.APP_TEMPLATE_FILE),
            os.path.join(app_name, app_name + constants.PY_EXT),
        )
        utils.cp(constants.ASSETS_TEMPLATE_DIR, constants.APP_ASSETS_DIR)

    # Install bun if it isn't already installed.
    if not os.path.exists(utils.get_bun_path()):
        typer.echo("Installing bun")
        os.system(constants.INSTALL_BUN)

    # Initialize the web directory.
    utils.rm(os.path.join(constants.WEB_TEMPLATE_DIR, constants.NODE_MODULES))
    utils.rm(os.path.join(constants.WEB_TEMPLATE_DIR, constants.PACKAGE_LOCK))
    utils.cp(constants.WEB_TEMPLATE_DIR, constants.WEB_DIR)


@cli.command()
def run(
    env: constants.Env = constants.Env.DEV.value,
    frontend: bool = True,
    backend: bool = True,
):
    """Run the app."""
    app = utils.get_app()
    if env == constants.Env.DEV:
        if frontend:
            utils.run_frontend(app)
        if backend:
            utils.run_backend(app)
    if env == constants.Env.PROD:
        if frontend:
            utils.run_frontend_prod(app)
        if backend:
            utils.run_backend_prod(app)


@cli.command()
def deploy(dry_run: bool = False):
    """Deploy the app to the hosting service."""
    # Get the app config.
    pcconfig = utils.get_config()
    pcconfig.API_HOST = utils.get_production_backend_url()

    # Compile the app in production mode.
    typer.echo("Compiling production app")
    app = utils.get_app()
    app.app.compile()

    # Export and zip up the frontend and backend.
    cmd = r"rm -rf .web/_static; cd .web && bun run export && cd _static  && zip -r ../../frontend.zip ./* && cd ../.. && zip -r backend.zip ./* -x .web/\* ./assets\* ./frontend.zip\* ./backend.zip\*"
    os.system(cmd)

    if dry_run:
        return

    data = {"userId": pcconfig.USERNAME, "projectId": pcconfig.APP_NAME}
    original_response = requests.get(
        "https://0v8g446r99.execute-api.us-west-1.amazonaws.com/prod/hosting-api",
        params=data,
    )
    response = original_response.json()
    print("response", response)
    frontend = response["frontend_resources_url"]
    backend = response["backend_resources_url"]

    # Upload the frontend and backend.
    with open("frontend.zip", "rb") as f:
        response = requests.put(frontend, data=f)

    with open("backend.zip", "rb") as f:
        response = requests.put(backend, data=f)


main = cli


if __name__ == "__main__":
    main()
